{$A+,B-,F+,G+,N+,P+,Q-,R-,T-,V+,X+}
{****************************************************************************
 * GrafX2 font unit - copyright (c) 1998-2000 by Stefan Goehler             *
 * Any bug reports and suggestions are welcome                              *
 * License agreement in gx2.txt - illegal copy without this file!           *
 * Complete package and new version always available here:                  *
 * http://sourcenet.home.pages.de                                           *
 ****************************************************************************}
{$DEFINE asmtext}      { Use assembler XFN routines - faster }
{$DEFINE ttf}          { Support for True Type Fonts         }

unit gx_font;

interface
uses grafx2,gx_glob,newfrontier,stream32{$IFDEF ttf},freetype{$ENDIF};

type
  pgxfont = ^tgxfont;
  tgxfont = object
    constructor init;
    destructor  close;

    function    height(s : string) : longint;
    function    width(s : string) : longint;
    procedure   write(var screen : tgrafx;x,y : longint;s : string;col : longint);
    function    maxcharheight : longint;
    function    maxcharwidth : longint;

    private
      fnalign         : tpoint;
      fnsizex,fnsizey : double;
  end;

  pxfn = ^txfn;
  txfn = object(tgxfont)
    constructor init(var stream : tstream32);
    destructor  close;

    function    height(s : string) : longint;
    function    width(s : string) : longint;
    procedure   write(var screen : tgrafx;x,y : longint;s : string;col : longint);
    function    maxcharheight : longint;
    function    maxcharwidth : longint;
    procedure   setfontattributes(xalign,yalign,flags : longint);
    procedure   getfontattributes(var xalign,yalign,flags : longint);
    procedure   setfontsize(size : double);
    function    getfontsize : double;
    procedure   setfontangle(angle : double);
    function    getfontangle : double;

    function    numofsubfonts : longint;

    function    savefont(var stream : tstream32) : word;
    private
      font          : tfont;
      fnangle       : double;
      fnflags       : longint;
      fnbmp8data    : tselector; { Color palette for BMP8 fonts }
  end;

{$IFDEF ttf}
	pcharpic = ^tcharpic;
	tcharpic = object
		xo,yo,advance  : integer;
    raster         : tt_raster_map;
		constructor init(const glyph: tt_glyph; const imetrics: tt_instance_metrics);
		destructor done;
	end;

  pttf = ^tttf;
  tttf = object(txfn)
    constructor init(fname : string);
    destructor  close;

    procedure   setfontsize(size : double);

    private
		charmap : array[byte] of pcharpic;
		glyph   : tt_glyph;

		face     : tt_face;
		instance : tt_instance;
		metrics  : tt_instance_metrics;

{    procedure   clearcharmap;}
		function    getcharpic(c : char) : pcharpic;
  end;
{$ENDIF}

var
  sysfont : txfn;

implementation

uses
	ttfile;

var
	openttf : word;


constructor tgxfont.init;
begin
end;

destructor tgxfont.close;
begin
end;

function tgxfont.height(s : string) : longint;
begin
  height := 0;
end;

function tgxfont.width(s : string) : longint;
begin
  width := 0;
end;

procedure tgxfont.write(var screen : tgrafx;x,y : longint;s : string;col : longint);
begin
end;

function tgxfont.maxcharheight : longint;
begin
  maxcharheight := 0;
end;

function tgxfont.maxcharwidth : longint;
begin
  maxcharwidth := 0;
end;


constructor txfn.init(var stream : tstream32);
var
  count : longint;
begin
  inherited init;
  if (stream.status <> stok) then fail;
  stream.read32(farptr(@font),sizeof(tfont));
  if (stream.status <> stok) then begin
    fail;
  end;
  if (font.id <> idfont)or(font.version <> $0101) then begin
    fail;
  end;
  if (font.size > memavail32) then begin
    fail;
  end;
  getmem32(font.data,(font.size+15) and not 7);
  count := stream.read32(ptr48(font.data,0),font.size);
  if (count <> font.size) then begin
    freemem32(font.data);
    fillbyte(farptr(@font),sizeof(tfont),0);
    fail;
  end;
  font.fontstart := 0;

  if (@font = nil)or(font.data = 0)or(font.id <> idfont) then exit;
  { If old font multifont, restore header }
  if (font.flags and fnmulti <> 0) then setfontsize(0);
  if (font.flags and fnvector <> 0) then
  begin
    fnsizex := 1;
    fnsizey := 1;
  end else
  begin
    fnsizex := 0;
    fnsizey := 0;
  end;
  if (font.usedcolors > 1) then getmem32(fnbmp8data,128*4);
end;

destructor txfn.close;
begin
  inherited close;
  if (font.usedcolors > 1) then freemem32(fnbmp8data);
  freemem32(font.data);
  if (font.id <> idfont)or(font.data = 0) then
    fillbyte(farptr(@font),sizeof(tfont),0);
  fillbyte(farptr(@font),sizeof(tfont),0);
end;

function txfn.height(s : string) : longint;
var
  i           : integer;
  ysize       : longint;
  mempos      : longint;
  sizey       : longint;
  subfont     : longint;
  arraystart  : longint;
  oldangle    : double;
begin
  ysize := 0;
  if (font.id <> idfont)or(font.data = 0)or(length(s) = 0) then { no font loaded }
  begin
    height := 0;
    exit;
  end;
  if (fnangle <> 0)and(fnangle <> 180) then
  begin
    oldangle := fnangle;
    fnangle  := 0;
    height   := width(s);
    fnangle  := oldangle;
    exit;
  end;
  if (font.id <> idfont)or(font.data = 0)or(length(s) = 0) then exit; { no font loaded }
  if (font.flags and fnfixed <> 0) then
  begin
    ysize := font.sizey;
  end else begin
    if (font.flags and fnvector = 0) then
    case font.bpp of
      1    : begin
               subfont := trunc(fnsizex);
               if subfont > font.numofsubfonts then
                subfont := font.numofsubfonts;
               arraystart := (font.endchar-font.startchar+1)*4
                             *subfont;
               for i := 1 to length(s) do begin
                 if (ord(s[i]) < font.startchar)or
                    (ord(s[i]) > font.endchar) then continue;

                 mempos := readmem32l(ptr48(font.data,
                                     longint(ord(s[i])-font.startchar) shl 2+
                                     arraystart));
                 sizey  := readmem32b(ptr48(font.data,mempos+1));
                 if (sizey > ysize) then ysize := sizey;
               end;
             end;
      2..32: begin
               subfont := trunc(fnsizex);
               if subfont > font.numofsubfonts then
                subfont := font.numofsubfonts;
               if (font.flags and fnanimate) <> 0 then subfont := 0;

               arraystart := (font.endchar-font.startchar+1)*4
                             *subfont;
               for i := 1 to length(s) do begin
                 if (ord(s[i]) < font.startchar)or
                    (ord(s[i]) > font.endchar) then continue;
                 mempos := readmem32l(ptr48(font.data,
                                     longint(ord(s[i])-font.startchar) shl 2+
                                     arraystart));
                 sizey  := readmem32l(ptr48(font.data,mempos+12))+
                           readmem32l(ptr48(font.data,mempos+4));
                 if (sizey > ysize) then ysize := sizey;
               end;
             end;
    end else { We have a vector font}
    begin
      subfont := trunc(fnsizex);
      if subfont > font.numofsubfonts then
       subfont := font.numofsubfonts;
      arraystart := (font.endchar-font.startchar+1)*4
                    *subfont;
      for i := 1 to length(s) do begin
        if (ord(s[i]) < font.startchar)or
           (ord(s[i]) > font.endchar) then continue;

        mempos := readmem32l(ptr48(font.data,
                            longint(ord(s[i])-font.startchar) shl 2+
                            arraystart));
        sizey  := readmem32b(ptr48(font.data,mempos+1));
        if (sizey > ysize) then ysize := sizey;
      end;
      ysize := round(ysize*(fnsizex));
    end;
  end;
  if (font.flags and fnanipos <> 0) then
   inc(ysize,font.randomy);
  height := ysize;
end;

function txfn.width(s : string) : longint;
type
 pvecchar = ^tvecchar;
 tvecchar = record
   sizex,sizey : byte;
   numoflines  : word;
   coordinates : array[0..0] of record
     code,x,y : byte;
   end;
 end;

var
  i           : integer;
  xsize,size  : word;
  mempos      : longint;
  subfont     : longint;
  arraystart  : longint;
  oldangle    : double;
  vec         : pvecchar;

begin
  xsize := 0;
  if (font.id <> idfont)or(font.data = 0)or(length(s) = 0) then { no font loaded }
  begin
    width := 0;
    exit;
  end;
  if (fnangle <> 0)and(fnangle <> 180) then
  begin
    oldangle := fnangle;
    fnangle  := 0;
    width    := height(s);
    fnangle  := oldangle;
    exit;
  end;
  if (font.id <> idfont)or(font.data = 0)or
     (length(s) = 0) then exit; { no font loaded }
  if (font.flags and fnfixed <> 0) then
  begin
    xsize := length(s)*(font.sizex+font.distance);
  end else
  begin
    if (font.flags and fnvector = 0) then
    case font.bpp of
      1    : begin
               subfont := trunc(fnsizex);
               if subfont > font.numofsubfonts then
                subfont := font.numofsubfonts;
               arraystart := (font.endchar-font.startchar+1)*4
                             *subfont;
               for i := 1 to length(s) do begin
                 if (ord(s[i]) < font.startchar)or
                    (ord(s[i]) > font.endchar) then
                    begin
                      inc(xsize,font.spacer+font.distance);
                      continue;
                    end;
                 mempos := readmem32l(ptr48(font.data,
                                     longint(ord(s[i])-font.startchar) shl 2+
                                     arraystart));
                 inc(xsize,readmem32b(ptr48(font.data,mempos))+
                           font.distance);
               end;
             end;
      2..32: begin
               subfont := trunc(fnsizex);
               if subfont > font.numofsubfonts then
                subfont := font.numofsubfonts;
               if (font.flags and fnanimate) <> 0 then subfont := 0;

               arraystart := (font.endchar-font.startchar+1)*4
                             *subfont;
               for i := 1 to length(s) do begin
                 if (ord(s[i]) < font.startchar)or
                    (ord(s[i]) > font.endchar) then
                    begin
                      inc(xsize,font.spacer+font.distance);
                      continue;
                    end;
                 mempos := readmem32l(ptr48(font.data,
                                     longint(ord(s[i])-font.startchar) shl 2+
                                     arraystart));
                 inc(xsize,readmem32l(ptr48(font.data,mempos))+
                           readmem32l(ptr48(font.data,mempos+8))+
                           font.distance);
               end;
             end;
    end else { We have a vector font}
    begin
{      arraystart := (font.endchar-font.startchar+1)*4;
      for i := 1 to length(s) do begin
        if (ord(s[i]) < font.startchar)or
           (ord(s[i]) > font.endchar) then
           begin
             inc(xsize,(font.spacer*trunc(fnsizex)) shr 1+font.distance);
             continue;
           end;
        mempos := readmem32l(ptr48(font.data,
                            longint(ord(s[i])-font.startchar) shl 2));
        inc(xsize,readmem32b(ptr48(font.data,mempos)));

      end;
      xsize := round(xsize*(fnsizex+1));
      xsize := xsize+font.distance*length(s);
    end;}

    for i := 1 to length(s) do begin
    if (font.startchar > ord(s[i]))or
       (font.endchar < ord(s[i])) then
      begin
        size := font.spacer;
      end else
      begin
        mempos := readmem32l(ptr48(font.data,(ord(s[i])-
                             font.startchar) shl 2));
        vec := ptr(font.data,mempos);
        size := round((vec^.sizex*(fnsizex)));
      end;
      inc(xsize,size+font.distance);
    end;

  end;
  end;
  if (font.flags and fnanipos <> 0) then
   inc(xsize,font.randomx);
  width := xsize-font.distance;
end;

procedure txfn.write(var screen : tgrafx;x,y : longint;s : string;col : longint);
var
  i,cx,cy     : integer;
  data        : longint;
  xpos        : longint;
  mempos      : longint;
  datapos     : longint;
  sizex,sizey : longint;
  addx,addy   : longint;
  crgb        : trgb;
  subfont     : longint;
  arraystart  : longint;
  datainc     : longint;
  scrpos      : longint;
  bpl         : longint;
  startchar,endchar : longint;
  distance    : longint;
  spacer      : longint;
  ymax,ystart : longint;
  posx,posy   : longint;
  cn,sn       : double;
  lnflags     : longint;

  drawpixel   : procedure(x,y,col : longint;var mode : tmode);

function outsymbol(x1,y1 : longint;c : char) : word;
type
 pvecchar = ^tvecchar;
 tvecchar = record
   sizex,sizey : byte;
   numoflines  : word;
   coordinates : array[0..0] of record
     code,x,y : byte;
   end;
 end;

var
  i,j         : word;
  xx,yy,mx,my : longint;
  vec         : pvecchar;
  x2,y2       : double;

begin
  if (font.startchar > ord(c))or
     (font.endchar < ord(c)) then
  begin
    outsymbol := font.spacer;
    exit;
  end;
  mempos := readmem32l(ptr48(font.data,(ord(c)-
                             font.startchar) shl 2));

  posx := round(cn*x1+sn*y1);
  posy := round(-sn*x1+cn*y1);

  vec := ptr(font.data,mempos);
  for j := 0 to Vec^.numoflines-1 do
   with vec^.coordinates[j] do begin
     if fnangle = 0 then
     begin
       mx := x1+round((x*(fnsizex)));
       my := y1+round((y*(fnsizey)));
     end else
     begin
       x2 := (x1+x*(fnsizex));
       y2 := (y1+y*(fnsizey));
       mx := round(cn*x2+sn*y2);
       my := round(-sn*x2+cn*y2);
     end;
     if (code = 1) then
     begin
       posx := mx;
       posy := my;
     end else
     begin
       screen.line(posx,posy,mx,my,col,lnflags);
       posx := mx;
       posy := my;
     end;
   end;
  outsymbol := round((vec^.sizex*(fnsizex)));
end;

begin
  if (font.id <> idfont)or(font.data = 0)or(length(s) = 0) then exit; { no font loaded }
  case fnalign.x of
    textcentered : x := x-width(s) shr 1;
    textright    : x := x-width(s);
  end;
  case fnalign.y of
    textcentered : y := y-height(s) shr 1;
    textbottom   : y := y-height(s);
  end;

  if (y > screen.maxy)or(x > screen.maxx) then exit;

  case (fnflags and (textadd or textsub)) of
    textdefault : drawpixel := screen.proc_pp;
    textadd     : drawpixel := screen.proc_addp;
    textsub     : drawpixel := screen.proc_subp;
  end;

  if (font.flags and fnvector = 0) then
  case font.bpp of
    1    : begin
             {$IFDEF ASMTEXT}
             if (y >= 0)and(x >= 0)and(x+width(s) <= screen.maxx)and
             (fnangle = 0)and
             (fnflags and (textadd or textsub) = 0)and
             (font.sizex <= 32) then
             begin
               subfont := trunc(fnsizex);
               if subfont > font.numofsubfonts then
                subfont := font.numofsubfonts;
               arraystart := (font.endchar-font.startchar+1)*4
                             *subfont;

               ymax      := screen.maxy-y+1;
               datainc   := font.bpl;
               bpl       := screen.mode.bpl;
               startchar := font.startchar;
               endchar   := font.endchar;
               distance  := font.distance;
               if (font.flags and fnfixed = 0)  then
                spacer := font.spacer*screen.mode.pixel+font.distance*screen.mode.pixel
                else spacer := font.sizex*screen.mode.pixel+font.distance*screen.mode.pixel;
               inc(x,screen.mode.actviewport.x1);
               inc(y,screen.mode.actviewport.y1);
               if (screen.mode.flags and mdlinear <> 0) then
               case screen.mode.bpp of
                 15,16:
                 asm
                          mov   i,1
                          les   di,screen
                   db 66h;mov   bx,x.word
                   db 66h;mov   ax,es:[di].tmode.bpl.word
                   db 66h;add   bx,bx
                   db 66h;mul   y.word
                   db 66h;add   ax,bx
                   db 66h;add   ax,es:[di].tmode.displayofs.word
                   db 66h;mov   si,ax
                          mov   ax,es:[di].tmode.writeptr
                   db 8Eh,0E8h {mov   gs,ax}
                          les   di,self
                          mov   ax,es:[di].font.tfont.data
                          mov   es,ax
                          jmp   @charloop
                   @notavailable:
                     db 66h;pop   si
                     db 66h;add   si,spacer.word
                            jmp   @loopend
                   @charloop:
                     db 66h;push  si

                            mov   di,i
                     db 66h;xor   ax,ax
                            mov   al,s[di].byte
                     db 66h;cmp   ax,endchar.word
                            jnbe  @notavailable
                     db 66h;sub   ax,startchar.word
                            jc    @notavailable

                     db 66h;shl   ax,2
                     db 66h;add   ax,arraystart.word
                     db 26h,67h,66h,8Bh,38h {mov   edi,es:[eax]}

                     db 66h;xor   dx,dx
                     db 66h;xor   cx,cx
                     db 26h,67h,8Ah,0Fh     {mov   cl,es:[edi]}
                     db 26h,67h,8Ah,57h,01h {mov   dl,es:[edi+1]}

                     db 66h;add   di,2
                     db 66h;mov   bx,col.word
                     db 66h;or    dx,dx
                            jz    @notavailable
                     db 66h;cmp   dx,ymax.word {check if y+charheight > maxy}
                            jbe   @ylp

                     db 66h;mov   dx,ymax.word
                     @ylp:

                       db 66h;push   cx
                       db 26h,67h,66h,8Bh,07h {mov   eax,es:[edi]}
                       @xlp:
                         db 66h;test  ax,1;dw 0000
                                jz    @nopix
                         db 65h,67h,89h,1Eh   {mov   gs:[esi],bx}
                                @nopix:
                         db 66h;shr   ax,1
                         db 66h;add   si,2
                         db 66h;dec   cx
                                jnz   @xlp

                       db 66h;pop   cx
                       db 66h;add   di,datainc.word
                       db 66h;add   si,bpl.word
                       db 66h;sub   si,cx
                       db 66h;sub   si,cx

                       db 66h;dec   dx
                              jnz   @ylp

                       db 66h;pop   si
                       db 66h;add   cx,distance.word
                       db 66h;add   cx,cx
                       db 66h;add   si,cx
                       @loopend:
                              mov   ax,i
                              inc   ax
                              mov   i,ax
                              cmp   al,s[0].byte
                              jna   @charloop
                   end;
                 24:
                 asm
                          mov   i,1
                          les   di,screen
                   db 66h;mov   bx,x.word
                   db 66h;mov   ax,es:[di].tmode.bpl.word
                   db 66h;mul   y.word
                   db 66h;add   ax,bx
                   db 66h;add   bx,bx
                   db 66h;add   ax,bx
                   db 66h;add   ax,es:[di].tmode.displayofs.word
                   db 66h;mov   si,ax
                          mov   ax,es:[di].tmode.writeptr
                   db 8Eh,0E8h {mov   gs,ax}
                          les   di,self
                          mov   ax,es:[di].font.tfont.data
                          mov   es,ax
                          jmp   @charloop
                   @notavailable:
                     db 66h;pop   si
                     db 66h;add   si,spacer.word
                            jmp   @loopend
                   @charloop:
                     db 66h;push  si

                            mov   di,i
                     db 66h;xor   ax,ax
                            mov   al,s[di].byte
                     db 66h;cmp   ax,endchar.word
                            jnbe  @notavailable
                     db 66h;sub   ax,startchar.word
                            jc    @notavailable

                     db 66h;shl   ax,2
                     db 66h;add   ax,arraystart.word
                     db 26h,67h,66h,8Bh,38h {mov   edi,es:[eax]}

                     db 66h;xor   dx,dx
                     db 66h;xor   cx,cx
                     db 26h,67h,8Ah,0Fh     {mov   cl,es:[edi]}
                     db 26h,67h,8Ah,57h,01h {mov   dl,es:[edi+1]}
                     db 66h;add   di,2
                     db 66h;mov   bx,col.word

                     db 66h;or    dx,dx
                            jz    @notavailable
                     db 66h;cmp   dx,ymax.word {check if y+charheight > maxy}
                            jbe   @ylp
                     db 66h;mov   dx,ymax.word
                     @ylp:

                       db 66h;push   cx
                       db 26h,67h,66h,8Bh,07h {mov   eax,es:[edi]}
                       @xlp:
                         db 66h;test  ax,1;dw 0
                                jz    @nopix
                         db 65h,67h,89h,1Eh     {mov   gs:[esi],bx}
                         db 66h;ror   ax,16
                         db 65h,67h,88h,5Eh,02h {mov   gs:[esi+2],bl}
                         db 66h;rol   ax,16
                                @nopix:
                         db 66h;shr   ax,1
                         db 66h;add   si,3
                         db 66h;dec   cx
                                jnz   @xlp

                       db 66h;pop   cx
                       db 66h;add   di,datainc.word
                       db 66h;add   si,bpl.word
                       db 66h;sub   si,cx
                       db 66h;sub   si,cx
                       db 66h;sub   si,cx

                       db 66h;dec   dx
                              jnz   @ylp

                       db 66h;pop   si
                       db 66h;add   cx,distance.word
                       db 66h;add   si,cx
                       db 66h;add   cx,cx
                       db 66h;add   si,cx
                       @loopend:
                              mov   ax,i
                              inc   ax
                              mov   i,ax
                              cmp   al,s[0].byte
                              jna   @charloop
                   end;
                 32:
                 asm
                          mov   i,1
                          les   di,screen
                   db 66h;mov   bx,x.word
                   db 66h;mov   ax,es:[di].tmode.bpl.word
                   db 66h;shl   bx,2
                   db 66h;mul   y.word
                   db 66h;add   ax,bx
                   db 66h;add   ax,es:[di].tmode.displayofs.word
                   db 66h;mov   si,ax
                          mov   ax,es:[di].tmode.writeptr
                   db 8Eh,0E8h {mov   gs,ax}
                          les   di,self
                          mov   ax,es:[di].font.tfont.data
                          mov   es,ax
                          jmp   @charloop
                   @notavailable:
                     db 66h;pop   si
                     db 66h;add   si,spacer.word
                            jmp   @loopend
                   @charloop:
                     db 66h;push  si

                            mov   di,i
                     db 66h;xor   ax,ax
                            mov   al,s[di].byte
                     db 66h;cmp   ax,endchar.word
                            jnbe  @notavailable
                     db 66h;sub   ax,startchar.word
                            jc    @notavailable

                     db 66h;shl   ax,2
                     db 66h;add   ax,arraystart.word
                     db 26h,67h,66h,8Bh,38h {mov   edi,es:[eax]}

                     db 66h;xor   dx,dx
                     db 66h;xor   cx,cx
                     db 26h,67h,8Ah,0Fh     {mov   cl,es:[edi]}
                     db 26h,67h,8Ah,57h,01h {mov   dl,es:[edi+1]}
                     db 66h;add   di,2
                     db 66h;mov   bx,col.word

                     db 66h;or    dx,dx
                            jz    @notavailable
                     db 66h;cmp   dx,ymax.word {check if y+charheight > maxy}
                            jbe   @ylp
                     db 66h;mov   dx,ymax.word
                     @ylp:

                       db 66h;push   cx
                       db 26h,67h,66h,8Bh,07h {mov   eax,es:[edi]}
                       @xlp:
                         db 66h;test  ax,1;dw 0
                                jz    @nopix
                         db 65h,67h,66h,89h,1Eh   {mov   gs:[esi],ebx}
                                @nopix:
                         db 66h;shr   ax,1
                         db 66h;add   si,4
                         db 66h;dec   cx
                                jnz   @xlp

                       db 66h;pop   cx
                       db 66h;add   di,datainc.word
                       db 66h;shl   cx,2
                       db 66h;add   si,bpl.word
                       db 66h;sub   si,cx
                       db 66h;shr   cx,2

                       db 66h;dec   dx
                              jnz   @ylp

                       db 66h;pop   si
                       db 66h;add   cx,distance.word
                       db 66h;shl   cx,2
                       db 66h;add   si,cx
                       @loopend:
                              mov   ax,i
                              inc   ax
                              mov   i,ax
                              cmp   al,s[0].byte
                              jna   @charloop
                   end;
                 end else
               case screen.mode.bpp of
                 15,16:
                 asm
                          mov   i,1
                          les   di,screen
                   db 66h;mov   bx,x.word
                   db 66h;mov   ax,es:[di].tmode.bpl.word
                   db 66h;mov   cx,y.word
                   db 66h;add   cx,es:[di].tmode.pageadd.word
                   db 66h;add   bx,bx
                   db 66h;mul   cx
                   db 66h;add   ax,bx
                   db 66h;mov   si,ax
                   db 66h;shr   ax,16
                          cmp   ax,lastbank
                          je    @noswitch
                          push  ax
                          call  setbank
                   @noswitch:
                          mov   ax,es:[di].tmode.writeptr
                   db 8Eh,0E8h {mov   gs,ax}
                          les   di,self
                          mov   ax,es:[di].font.tfont.data
                          mov   es,ax
                          jmp   @charloop
                   @notavailable:
                     db 66h;pop   si
                     db 66h;add   si,spacer.word
                            jmp   @loopend
                   @charloop:
                     db 66h;push  si

                            mov   di,i
                     db 66h;xor   ax,ax
                            mov   al,s[di].byte
                     db 66h;cmp   ax,endchar.word
                            jnbe  @notavailable
                     db 66h;sub   ax,startchar.word
                            jc    @notavailable

                     db 66h;shl   ax,2
                     db 66h;add   ax,arraystart.word
                     db 26h,67h,66h,8Bh,38h {mov   edi,es:[eax]}

                     db 66h;xor   dx,dx
                     db 66h;xor   cx,cx
                     db 26h,67h,8Ah,0Fh     {mov   cl,es:[edi]}
                     db 26h,67h,8Ah,57h,01h {mov   dl,es:[edi+1]}
                     db 66h;add   di,2
                     db 66h;mov   bx,col.word

                     db 66h;or    dx,dx
                            jz    @notavailable
                     db 66h;cmp   dx,ymax.word {check if y+charheight > maxy}
                            jbe   @ylp
                     db 66h;mov   dx,ymax.word
                     @ylp:

                       db 66h;push   cx
                       db 26h,67h,66h,8Bh,07h {mov   eax,es:[edi]}
                       @xlp:
                         db 66h;test  ax,1;dw 0
                                jz    @nopix
                         db 65h,89h,1Ch   {mov   gs:[si],bx}
                                @nopix:
                         db 66h;shr   ax,1
                                add   si,2
                                jc    @switch
                                @continue:
                         db 66h;dec   cx
                                jnz   @xlp

                       db 66h;pop   cx
                       db 66h;add   di,datainc.word

                       db 66h;mov   ax,bpl.word
                       db 66h;sub   ax,cx
                       db 66h;sub   ax,cx
                              add   si,ax
                              jnc   @noswitch2
                       db 66h,81h,0C6h,00h,00h,01h,00h {add  esi,10000h}
                              call  incbank
                       @noswitch2:

                       db 66h;dec   dx
                              jnz   @ylp

                       db 66h;pop   si
                       db 66h;add   cx,distance.word
                       db 66h;add   cx,cx
                       db 66h;add   si,cx
                       db 66h;mov   ax,si
                       db 66h;shr   ax,16
                              cmp   ax,lastbank
                              je    @loopend
                              push  ax
                              call  setbank
                       @loopend:
                              mov   ax,i
                              inc   ax
                              mov   i,ax
                              cmp   al,s[0].byte
                              jna   @charloop
                              jmp   @end
                     @switch:
                     db 66h,81h,0C6h,00h,00h,01h,00h {add  esi,10000h}
                            call  incbank
                            jmp   @continue
                     @end:
                 end;
                 32:
                 asm
                          mov   i,1
                          les   di,screen
                   db 66h;mov   bx,x.word
                   db 66h;mov   ax,es:[di].tmode.bpl.word
                   db 66h;mov   cx,y.word
                   db 66h;add   cx,es:[di].tmode.pageadd.word
                   db 66h;shl   bx,2
                   db 66h;mul   cx
                   db 66h;add   ax,bx
                   db 66h;mov   si,ax
                   db 66h;shr   ax,16
                          cmp   ax,lastbank
                          je    @noswitch
                          push  ax
                          call  setbank
                   @noswitch:
                          mov   ax,es:[di].tmode.writeptr
                   db 8Eh,0E8h {mov   gs,ax}
                          les   di,self
                          mov   ax,es:[di].font.tfont.data
                          mov   es,ax
                          jmp   @charloop
                   @notavailable:
                     db 66h;pop   si
                     db 66h;add   si,spacer.word
                            jmp   @loopend
                   @charloop:
                     db 66h;push  si

                            mov   di,i
                     db 66h;xor   ax,ax
                            mov   al,s[di].byte
                     db 66h;cmp   ax,endchar.word
                            jnbe  @notavailable
                     db 66h;sub   ax,startchar.word
                            jc    @notavailable

                     db 66h;shl   ax,2
                     db 66h;add   ax,arraystart.word
                     db 26h,67h,66h,8Bh,38h {mov   edi,es:[eax]}

                     db 66h;xor   dx,dx
                     db 66h;xor   cx,cx
                     db 26h,67h,8Ah,0Fh     {mov   cl,es:[edi]}
                     db 26h,67h,8Ah,57h,01h {mov   dl,es:[edi+1]}
                     db 66h;add   di,2
                     db 66h;mov   bx,col.word

                     db 66h;or    dx,dx
                            jz    @notavailable
                     db 66h;cmp   dx,ymax.word {check if y+charheight > maxy}
                            jbe   @ylp
                     db 66h;mov   dx,ymax.word
                     @ylp:

                       db 66h;push   cx
                       db 26h,67h,66h,8Bh,07h {mov   eax,es:[edi]}
                       @xlp:
                         db 66h;test  ax,1;dw 0
                                jz    @nopix
                         db 65h,66h,89h,1Ch   {mov   gs:[si],ebx}
                                @nopix:
                         db 66h;shr   ax,1
                                add   si,4
                                jc    @switch
                                @continue:
                         db 66h;dec   cx
                                jnz   @xlp

                       db 66h;pop   cx
                       db 66h;add   di,datainc.word

                       db 66h;shl   cx,2
                       db 66h;mov   ax,bpl.word
                       db 66h;sub   ax,cx
                       db 66h;shr   cx,2
                              add   si,ax
                              jnc   @noswitch2
                       db 66h,81h,0C6h,00h,00h,01h,00h {add  esi,10000h}
                              call  incbank
                       @noswitch2:

                       db 66h;dec   dx
                              jnz   @ylp

                       db 66h;pop   si
                       db 66h;add   cx,distance.word
                       db 66h;shl   cx,2
                       db 66h;add   si,cx
                       db 66h;mov   ax,si
                       db 66h;shr   ax,16
                              cmp   ax,lastbank
                              je    @loopend
                              push  ax
                              call  setbank
                       @loopend:
                              mov   ax,i
                              inc   ax
                              mov   i,ax
                              cmp   al,s[0].byte
                              jna   @charloop
                              jmp   @end
                     @switch:
                     db 66h,81h,0C6h,00h,00h,01h,00h {add  esi,10000h}
                            call  incbank
                            jmp   @continue
                     @end:
                 end;
               end;
             end else
             {$ENDIF}
             begin
               if (fnangle = 90) then begin
                 xpos := y;
                 subfont := trunc(fnsizex);
                 if subfont > font.numofsubfonts then
                  subfont := font.numofsubfonts;
                 arraystart := (font.endchar-font.startchar+1)*4
                               *subfont;

                 for i := length(s) downto 1 do begin
                   if (ord(s[i]) < font.startchar)or
                      (ord(s[i]) > font.endchar) then
                      begin
                        if (font.flags and fnfixed = 0)  then
                        inc(xpos,font.spacer+font.distance)
                        else inc(xpos,font.sizex+font.distance);
                        continue;
                      end;
                   mempos := readmem32l(ptr48(font.data,
                                       longint(ord(s[i])-font.startchar) shl 2+
                                       arraystart));
                   sizex  := readmem32b(ptr48(font.data,mempos));
                   sizey  := readmem32b(ptr48(font.data,mempos+1));
                   inc(mempos,2);
                   for cy := 0 to sizey-1 do begin
                     datapos := 0;
                     data := readmem32l(ptr48(font.data,mempos));
                     for cx := 0 to sizex-1 do begin
                       if (cx > 31)and(cx and 31 = 0) then
                       begin { load new pixels }
                         inc(datapos,4);
                         data := readmem32l(ptr48(font.data,mempos+datapos));
                       end;
                       if data and (longint(1) shl (cx and 31)) <> 0 then
                         drawpixel(x+cy,xpos+sizex-cx,col,screen.mode);
                     end;
                     inc(mempos,font.bpl);
                   end;
                   inc(xpos,sizex+font.distance);
                 end;
               end else
               if (fnangle = 270) then begin
                 xpos := y;
                 subfont := trunc(fnsizex);
                 if subfont > font.numofsubfonts then
                  subfont := font.numofsubfonts;
                 arraystart := (font.endchar-font.startchar+1)*4
                               *subfont;

                 for i := 1 to length(s) do begin
                   if (ord(s[i]) < font.startchar)or
                      (ord(s[i]) > font.endchar) then
                      begin
                        if (font.flags and fnfixed = 0)  then
                        inc(xpos,font.spacer+font.distance)
                        else inc(xpos,font.sizex+font.distance);
                        continue;
                      end;
                   mempos := readmem32l(ptr48(font.data,
                                       longint(ord(s[i])-font.startchar) shl 2+
                                       arraystart));
                   sizex  := readmem32b(ptr48(font.data,mempos));
                   sizey  := readmem32b(ptr48(font.data,mempos+1));
                   inc(mempos,2);
                   for cy := 0 to sizey-1 do begin
                     datapos := 0;
                     data := readmem32l(ptr48(font.data,mempos));
                     for cx := 0 to sizex-1 do begin
                       if (cx > 31)and(cx and 31 = 0) then
                       begin { load new pixels }
                         inc(datapos,4);
                         data := readmem32l(ptr48(font.data,mempos+datapos));
                       end;
                       if data and (longint(1) shl (cx and 31)) <> 0 then
                         drawpixel(x+sizey-cy,xpos+cx,col,screen.mode);
                     end;
                     inc(mempos,font.bpl);
                   end;
                   inc(xpos,sizex+font.distance);
                 end;
               end else
               begin
                 xpos := x;
                 subfont := trunc(fnsizex);
                 if subfont > font.numofsubfonts then
                  subfont := font.numofsubfonts;
                 arraystart := (font.endchar-font.startchar+1)*4
                               *subfont;

                 for i := 1 to length(s) do begin
                   if (ord(s[i]) < font.startchar)or
                      (ord(s[i]) > font.endchar) then
                      begin
                        if (font.flags and fnfixed = 0)  then
                        inc(xpos,font.spacer+font.distance)
                        else inc(xpos,font.sizex+font.distance);
                        continue;
                      end;
                   mempos := readmem32l(ptr48(font.data,
                                       longint(ord(s[i])-font.startchar) shl 2+
                                       arraystart));
                   sizex  := readmem32b(ptr48(font.data,mempos));
                   sizey  := readmem32b(ptr48(font.data,mempos+1));
                   inc(mempos,2);
                   for cy := 0 to sizey-1 do begin
                     datapos := 0;
                     data := readmem32l(ptr48(font.data,mempos));
                     for cx := 0 to sizex-1 do begin
                       if (cx > 31)and(cx and 31 = 0) then
                       begin { load new pixels }
                         inc(datapos,4);
                         data := readmem32l(ptr48(font.data,mempos+datapos));
                       end;
                       if data and (longint(1) shl (cx and 31)) <> 0 then
                         drawpixel(xpos+cx,y+cy,col,screen.mode);
                     end;
                     inc(mempos,font.bpl);
                   end;
                   inc(xpos,sizex+font.distance);
                 end;
               end;
             end;
           end;
    2..7 : begin
             if col <> $FFFFFFFF then begin
               screen.dergb(col,crgb);
               for i := 1 to font.usedcolors do
                 writemem32l(ptr48(fnbmp8data,i shl 2),
                             screen.rgb(longint(crgb.r)*i div pred(font.usedcolors),
                             longint(crgb.g)*i div pred(font.usedcolors),
                             longint(crgb.b)*i div pred(font.usedcolors)));
             end;
             if (font.flags and fnanipos) <> 0 then begin
               randarray_word(gfxbufptr,length(s)+1,font.randomx);
               randarray_word(ptr48(gfxbufsel,512),length(s)+1,font.randomy);
             end;
             xpos := x;
             subfont := trunc(fnsizex);
             if subfont > font.numofsubfonts then
              subfont := font.numofsubfonts;
             if (font.flags and fnanimate) <> 0 then subfont := 0;

             arraystart := (font.endchar-font.startchar+1)*4
                           *subfont;
             for i := 1 to length(s) do begin
               if (ord(s[i]) < font.startchar)or
                  (ord(s[i]) > font.endchar) then
                  begin
                    if (font.flags and fnfixed = 0)  then
                    inc(xpos,font.spacer+font.distance)
                    else inc(xpos,font.sizex+font.distance);
                    continue;
                  end;
               mempos := readmem32l(ptr48(font.data,
                                   longint(ord(s[i])-font.startchar) shl 2+
                                   arraystart));
               addx   := readmem32l(ptr48(font.data,mempos));
               addy   := readmem32l(ptr48(font.data,mempos+4));
               if (font.flags and fnanipos <> 0) then begin
                 inc(addx,memw[gfxbufsel:i shl 1]);
                 inc(addy,memw[gfxbufsel:i shl 1+512]);
               end;
               sizex  := readmem32l(ptr48(font.data,mempos+8));
               sizey  := readmem32l(ptr48(font.data,mempos+12));
               inc(mempos,16);
               inc(xpos,addx);
               for cy := 0 to sizey-1 do
               for cx := 0 to sizex-1 do begin
                 data := readmem32b(ptr48(font.data,mempos));
                 inc(mempos);
                 if data and $80 = 0 then
                 drawpixel(xpos+cx,y+addy+cy,readmem32l(ptr48(fnbmp8data,data shl 2)),screen.mode);
               end;
               if (font.flags and fnanipos <> 0) then
                 dec(xpos,memw[gfxbufsel:i shl 1]);
               if (font.flags and fnfixed <> 0) then
                 inc(xpos,font.sizex+font.distance)
                 else inc(xpos,sizex+font.distance);
             end;
           end;
  end else {font is a vector font}
  begin
    sn:=sin(pi*fnangle/180);
    cn:=cos(pi*fnangle/180);
    xpos := x;
    if (fnflags and textantialias <> 0) then
      lnflags := lnantialias else lnflags := lndefault;
    for i := 1 to length(s) do begin
       inc(xpos,outsymbol(xpos,y,{fill,}s[i])+font.distance);
    end;
  end;
  if (fnflags and textunderline <> 0) then begin
    case round(fnangle) of
      0   : screen.hline(x-1,y+font.baseline+2,width(s)+2,col);
      90  : screen.vline(x+font.baseline+2,y-1,height(s)+2,col);
      270 : screen.vline(x-2,y-1,height(s)+2,col);
    end;
  end;
  if (fnflags and textstrikeout <> 0) then begin
    case round(fnangle) of
      0   : screen.hline(x-1,y+height('abcde') shr 1,width(s)+2,col);
      90  : screen.vline(x+width('abcde') shr 1,y-1,height(s)+2,col);
      270 : screen.vline(x+width('abcde')-width('abcde') shr 1,y,height(s)+2,col);
    end;
  end;
end;

function txfn.maxcharheight : longint;
var
  h : longint;
begin
  if (font.id <> idfont) then exit;
  if (font.flags and fnvector = 0) then
    h := font.sizey else
    h := round(font.sizey*fnsizey);

  if (font.flags and fnanipos <> 0) then
    inc(h,font.randomy);
  maxcharheight := h;
end;

function txfn.maxcharwidth : longint;
var
  w : longint;
begin
  if (font.id <> idfont) then exit;
  if (font.flags and fnvector = 0) then
    w := font.sizex else
    w := round(font.sizex*fnsizex);

  if (font.flags and fnanipos <> 0) then
    inc(w,font.randomx);
  maxcharwidth := w;
end;

procedure txfn.setfontattributes(xalign,yalign,flags : longint);
begin
  fnalign.x := xalign;
  fnalign.y := yalign;
  fnflags   := flags;
end;

procedure txfn.getfontattributes(var xalign,yalign,flags : longint);
begin
  xalign := fnalign.x;
  yalign := fnalign.y;
  flags  := fnflags;
end;

procedure txfn.setfontsize(size : double);
begin
  if (font.numofsubfonts <> 0) then begin
    if (size > font.numofsubfonts) then
      size := font.numofsubfonts;
      move2(ptr48(font.data,(font.endchar-
                  font.startchar+1)*4*
                  succ(font.numofsubfonts)+trunc(size)*sizeof(tsubfont)),
            farptr(@font.fontstart),sizeof(tsubfont));
  end;
  fnsizex := size;
  fnsizey := size;
end;

function txfn.getfontsize : double;
begin
  getfontsize := fnsizex;
end;

procedure txfn.setfontangle(angle : double);
begin
  fnangle := angle;
end;

function txfn.getfontangle : double;
begin
  getfontangle := fnangle;
end;

function txfn.numofsubfonts : longint;
begin
  numofsubfonts := font.numofsubfonts;
end;

function txfn.savefont(var stream : tstream32) : word;
var
  i    : integer;
  l    : longint;
  data : tselector;
begin
  savefont := errfileerror;
  if (stream.status <> stok) then exit;
  savefont  := errnodiskmem;
  stream.write32(farptr(@font),sizeof(tfont));
  if (stream.status <> stok) then exit;
  data      := font.data;
  font.data := 0;
  stream.write32(ptr48(data,font.fontstart),font.size);
  if (stream.status <> stok) then exit;
  font.data := data;
  savefont := errnoerror;
end;

{$IFDEF ttf}
constructor tcharpic.init(const glyph : tt_glyph;const imetrics: tt_instance_metrics);
var
  metrics : tt_glyph_metrics;
begin
  tt_get_glyph_metrics(glyph,metrics);

  with metrics,bbox,raster do
  begin
    width  := (xmax-xmin+63) div 64+1;
    rows   := (ymax-ymin+63) div 64+1;
    flow   := tt_flow_down;
    cols   := (width+7) div 8;
    size   := cols*rows;
    getmem(buffer,size);
    fillchar(buffer^,size,0);
    tt_get_glyph_bitmap(glyph,raster,-(xmin and (not 63)),-(ymin and (not 63)));
    xo     := (xmin+63) div 64;
    yo     := -(ymax+63) div 64+((imetrics.y_ppem*3) shr 2);
  end;
  advance := metrics.advance div 64+1;
end;

destructor tcharpic.done;
begin
  if assigned(raster.buffer) then
    freemem(raster.buffer,raster.size);
end;

constructor tttf.init(fname : string);
begin
  if (openttf = 0) then
    tt_init_freetype;
  inc(openttf);
  if (tt_open_face(fname,face) <> 0) then
    fail;
  tt_new_glyph(face,glyph);
  tt_new_instance(face,instance);
  tt_set_instance_resolutions(instance,96,96);
  setfontsize(7);
end;

destructor tttf.close;
begin
{	clearcharmap;}
  tt_close_face(face);
  dec(openttf);
  if (openttf = 0) then
    tt_done_freetype;
  inherited close;
end;

procedure tttf.setfontsize(size : double);
var
  i,j,x,y : longint;
  mempos  : longint;
  b       : byte;
  dat     : array[0..31] of byte;
  last    : integer;
  sx      : word;

begin
  if (font.id = idfont) then
  begin
    freemem32(font.data);
    fillbyte(farptr(@font),sizeof(tfont),0);
  end;

  tt_set_instance_charsize(instance,round(size*64));
  tt_get_instance_metrics(instance,metrics);

  with font do begin
    id        := 'Fnt2';
    version   := $0101;
    fonttype  := 'BMP1';
    bpp       := 1;
    sizex     := 0;
    sizey     := 0;
    startchar := 29;
    endchar   := 255;
    size      := 0;
    flags     := 0;
    numofsubfonts := 0;
    distance  := 1;
    fontstart := 0;
    usedcolors:= 2;
  end;

  last := 0;
  for i := font.startchar to font.endchar do
	with getcharpic(chr(i))^ do if (raster.buffer <> nil) then
    if last > yo then last := yo;

  last := -last;
  for i := font.startchar to font.endchar do
	with charmap[i]^ do if (raster.buffer <> nil) then begin
    if (raster.width+xo > font.sizex) then font.sizex := raster.width+xo;
    if (raster.rows+yo+last > font.sizey) then font.sizey := raster.rows+yo+last;
	end;
  font.bpl := (font.sizex+7) shr 3;
  with font do
  size := longint(2+4+sizey*font.bpl)*(endchar-startchar+1);
  getmem32(font.data,font.size);
  fillchar32(ptr48(font.data,0),font.size,0);
  font.spacer := font.sizex shr 1;

  for i := font.startchar to font.endchar do
	with charmap[i]^ do begin
    with font do
    mempos := (2+sizey*font.bpl)*(i-startchar)+4*(endchar-startchar+1);
    writemem32l(ptr48(font.data,(i-font.startchar)*4),mempos);

    if (raster.buffer <> nil) then
    begin
      sx := raster.width+xo;
      if sx > 0 then dec(sx);
      if (i = 32) {space} then sx := advance;
      writemem32w(ptr48(font.data,mempos),(sx and $FF)+word(raster.rows+yo+last) shl 8);
      inc(mempos,2+(yo+last)*font.bpl);
      for y := 0 to raster.rows-1 do begin
        for j := 0 to font.bpl-1 do begin
          dat[j] := 0;
          b := mem[seg(raster.buffer^):ofs(raster.buffer^)+j+y*raster.cols];
          for x := 0 to 7 do
            inc(dat[j],(((b shr (7-x)) and 1) shl x));
        end;
        move2(farptr(@dat),ptr48(font.data,mempos),raster.cols);
        inc(mempos,font.bpl);
      end;
      dispose(charmap[i],done);
      charmap[i] := nil;
    end else writemem32w(ptr48(font.data,mempos),0);
	end;
  font.baseline := font.sizey;
end;

function tttf.getcharpic(c : char): pcharpic;
begin
  if not assigned(charmap[byte(c)]) then
  begin
    tt_load_glyph(instance,glyph,ord(c)-29,tt_load_default);
    new(charmap[byte(c)],init(glyph,metrics));
  end;
  getcharpic := charmap[byte(c)];
end;
{$ENDIF}

var
  stream : pmemstream32;

var
  oldexit : pointer;

procedure font_exit;far;
begin
  sysfont.close;
  exitproc := oldexit;
end;

begin
  new(stream,init(farptr(@sysfont2),3712));
  sysfont.init(stream^);
  dispose(stream);

  oldexit  := exitproc;
  exitproc := @font_exit;
end.