{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}
unit ttflib;

interface

uses
	grafx2,FreeType; {TrueType support thanks to FreeType, see HTTP://FREETYPE.ORG}

type
	PCharPic=^TCharPic;
	TCharPic=object
		Width,Height,Advance,size,cols,xo,yo: Integer;
		Data: pointer;
		constructor Init(const Glyph: TT_Glyph; const IMetrics: TT_Instance_Metrics);
		destructor Done;
	end;
	PTrueTypeFont=^TTrueTypeFont;
	TTrueTypeFont=object{(TFont)}
	private
		CharMap: array[Char] of PCharPic;
		Glyph: TT_Glyph;
	public
		Face: TT_Face;
		Instance: TT_Instance;
		Metrics: TT_Instance_Metrics;
		constructor Init(const FName: string);
		procedure SetSize(ASize: Integer);
		function Width(s : string): longint; virtual;
		function Height: Integer; virtual;
    procedure write(var screen : tgrafx;s : string;px,py,col : longint); virtual;
		function GetCharPic(C: Char): PCharPic;
		procedure ClearCharMap;
		destructor Done; virtual;
	end;

implementation

uses
	TTFile;

var
	OpenTTF: Integer;

function Minimum(X,Y: Integer): Integer;
inline($58/         {pop ax}
		 $5B/           {pop bx}
		 $39/$D8/       {cmp ax,bx}
		 $7E/$02/       {jge  $+4}
		 $89/$D8);      {mov ax,bx}

constructor TTrueTypeFont.Init(const FName: string);
begin
	if OpenTTF=0 then
		TT_Init_FreeType;
	Inc(OpenTTF);
	if TT_Open_Face(FName,Face)<>0 then
		Fail;
	TT_New_Glyph(Face,Glyph);
	TT_New_Instance(Face,Instance);
	TT_Set_Instance_Resolutions(Instance,96,96);
	SetSize(7);
end;

procedure TTrueTypeFont.SetSize(ASize: Integer);
begin
	ClearCharMap;
	TT_Set_Instance_CharSize(Instance,ASize*64);
	TT_Get_Instance_Metrics(Instance,Metrics);
end;

function TTrueTypeFont.Width(s : string): longint;
var
  w,i : longint;
begin
  w := 0;
  for i := 1 to length(s) do inc(w,getcharpic(s[i])^.advance);
  width := w;
end;

function TTrueTypeFont.Height: Integer;
begin
	Height:=Metrics.Y_PPEM;
end;

procedure TTrueTypeFont.write(var screen : tgrafx;s : string;px,py,col : longint);
var
  i,j : integer;
  x,y : longint;
type
	TSetArray=array[0..1023] of set of 0..7;
	PSetArray=^TSetArray;
begin
  for i := 1 to length(s) do
	with getcharpic(s[i])^ do begin
  if (data <> nil) then
    for y := 0 to height-1 do
    for x := 0 to width-1 do
{    if mem[seg(data^):ofs(data^)+x shr 3+y*cols shr 3] and
       (x and 3) <> 0 then}
if ((7-(X and 7)) in PSetArray(data)^[X shr 3+Y*Cols]) then
       screen.putpixel(x+px+xo,y+py+yo,col);
    inc(px,advance);
	end;
end;

function TTrueTypeFont.GetCharPic(C: Char): PCharPic;
begin
	if not Assigned(CharMap[C]) then begin
		TT_Load_Glyph(Instance,Glyph,Ord(C)-29,TT_Load_Default);
		New(CharMap[C],Init(Glyph,Metrics));
	end;
	GetCharPic:=CharMap[C];
end;

procedure TTrueTypeFont.ClearCharMap;
var
	C: Char;
begin
	for C:=Low(Char) to High(Char) do
		if Assigned(CharMap[C]) then begin
			Dispose(CharMap[C],Done);
			CharMap[C]:=nil;
		end;
end;

destructor TTrueTypeFont.Done;
begin
	ClearCharMap;
	TT_Close_Face(Face);
	Dec(OpenTTF);
	if OpenTTF=0 then
		TT_Done_FreeType;
end;

constructor TCharPic.Init(const Glyph: TT_Glyph; const IMetrics: TT_Instance_Metrics);
type
	TSetArray=array[0..1023] of set of 0..7;
	PSetArray=^TSetArray;
var
	X,YC,I,TempCount : Integer;
	Metrics: TT_Glyph_Metrics;
	Raster: TT_Raster_Map;
begin
	TT_Get_Glyph_Metrics(Glyph,Metrics);
	with Metrics,BBox,Raster do begin
		Width:=(XMax-XMin+63) div 64+1;
		Rows:=(YMax-YMin+63) div 64+1;
		Flow:=TT_Flow_Down;
		Cols:=(Width+7) div 8;
		Size:=Cols*Rows;
		GetMem(Buffer,Size);
		FillChar(Buffer^,Size,0);
		TT_Get_Glyph_Bitmap(Glyph,Raster,-(XMin and (not 63)),-(YMin and (not 63)));
		XO:=(XMin+63) div 64;
		YO:=-(YMax+63) div 64+((IMetrics.Y_PPEM*3) shr 2);
		self.Width  := Width;
		self.Height := Rows;
    self.cols   := cols;
	end;
	Advance:=Metrics.Advance div 64+1;
  size := raster.size;
  getmem(data,size);
  move(raster.buffer^,data^,size);
	with Raster do
		FreeMem(Buffer,Size);
end;

destructor TCharPic.Done;
begin
	if Assigned(Data) then
		FreeMem(Data,size);
end;

end.